@@ -1,6 +1,5 @@
 import torch as T
 import torch.nn as nn
-import torch.nn.functional as F
 from lpd.extensions.custom_layers import Dense
 
 def examples_data_generator(N, D_in, D_out, binary_out=False):
@@ -13,9 +12,16 @@ def examples_data_generator(N, D_in, D_out, binary_out=False):
     while True:
         yield x, y #YIELD THE SAME X,y every time
 
+def examples_prediction_data_generator(origin_data_loader, steps):
+    for x,y in origin_data_loader:
+        steps -= 1
+        yield x
+        if steps == 0:
+            break
+
 
 def get_basic_model(D_in, H, D_out):
     return nn.Sequential(
-                            Dense(D_in, H, use_bias=True, activation=F.relu),
+                            Dense(D_in, H, use_bias=True, activation=nn.ReLU()),
                             Dense(H, D_out, use_bias=True, activation=None)
                         )
\ No newline at end of file
